import collections
import copy
import logging
import os
import pathlib
import re
import xml.etree.ElementTree as ET

import numpy as np
import cosa.utils
from cosa.utils import OrderedDefaultdict

logger = logging.getLogger(__name__)
logger.setLevel(logging.NOTSET)  # capture everything
# logger.setLevel(logging.) # capture everything
logger.disabled = True

all_var_names = ["Weights", "Inputs", "Outputs"]

var_dep = {
    "Weights": [],
    "Inputs": [],
    "Outputs": [],  # Partial sum does not dependent on any var
    "Outputs_Update": ["Weights", "Inputs", "Outputs"]
# Must wait until last output is written to the buf? maybe delete Outputs
}

var_bits = {
    "Weights": 8,
    "Inputs": 8,
    "Outputs": 24,
    "Outputs_Store": 24
}

dim_var = {
    0: ['Weights'],  # R
    1: ['Weights'],  # S
    2: ['Outputs', 'Inputs'],  # P
    3: ['Outputs', 'Inputs'],  # Q
    4: ['Weights', 'Inputs'],  # C
    5: ['Weights', 'Outputs'],  # K
    6: ['Inputs', 'Outputs'],  # N
}


def xml2dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = OrderedDefaultdict(list)
        for dc in map(xml2dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v
                     for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update(('@' + k, v)
                        for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d


def addr_to_str(addr):
    addr_arr = [str(a) for a in addr]
    return "_".join(addr_arr)


# Construct addrs dict from timeloop info
# Create a dictionary for different addresses, i would be the PE idx 
def construct_addrs_dict(buf):
    addrs = collections.OrderedDict()

    for i, addr in enumerate(buf):
        addr_key = addr_to_str(addr)
        if addr_key in addrs.keys():
            addrs[addr_key].append(i)
        else:
            addrs[addr_key] = [i]
    return addrs


def get_summary_info(stats_file):
    summary = {}
    with open(stats_file, 'r') as f:
        lines = f.readlines()
    for line in lines:
        # m = re.match(r"Energy: (.*) uJ", line)
        m = re.match(r"Total topology energy: (.*) pJ", line)
        if m:
            energy = m.group(1)
            summary['energy'] = float(energy)
        else:
            m = re.match(r"Max topology cycles: (.*)", line)
            if m:
                cycle = m.group(1)
                summary['cycle'] = int(cycle)
    return summary


"""

Returns: 
subnest_info = {
    subnest = {
        'Registers': [{'end': '1', 'spacetime_dimension': '0', '@class_id': '11', '@tracking_level': '0', 'start': '0', 'stride': '1', '@version': '0', 'dimension': '6'}, ...],  # arrays of subnest info 
        
    },
    bufsize = {
        'Registers': ('Registers', [0, 1, 0]), # buffered size for Weights, Inputs, Outputs
        ...
    }

}

"""


def get_subnest_info(xml_file):
    logger.info("================ Parse Subnest Info from {0} ================".format(xml_file))
    xml_path = pathlib.Path(xml_file)
    stats_file = xml_path.parent / xml_path.name.replace('map+stats.xml', 'stats.txt')
    summary_info = get_summary_info(stats_file)

    subnest_info = {}
    subnest_info['energy'] = summary_info['energy']
    subnest_info['cycle'] = summary_info['cycle']
    tree = ET.parse(xml_file);
    root = tree.getroot()

    timeloop_dict = xml2dict(root)
    # print(timeloop_dict)

    # print(timeloop_dict['boost_serialization']['best_mapped_engine']['topology_']['levels_']['item'])
    arch = timeloop_dict['boost_serialization']['engine']['topology_']['levels_']['item']
    arith = arch[0]
    # print(arith['px']['@class_name'])
    subnest_info['pe_cycle'] = int(arith['px']['cycles_'])
    subnest_info['pe_energy'] = float(arith['px']['energy_'])

    bufsize = collections.OrderedDict()
    subnest = collections.OrderedDict()

    for buf in arch[1:]:
        level_name = buf['px']['specs_']['LevelSpecs']['level_name']
        bufsize_arr = buf['px']['stats_']['utilized_capacity']['PerDataSpace']['item']
        bufsize_arr = [int(size) for size in bufsize_arr]

        bufsize[level_name] = bufsize_arr  # weight, input, output
        subnest_attr = buf['px']['subnest_']['item']
        if not isinstance(subnest_attr, list):
            subnest_attr = [subnest_attr]

        subnest[level_name] = subnest_attr

    subnest_info['bufsize'] = bufsize
    subnest_info['subnest'] = subnest
    return subnest_info


def get_num_spatial_cores(subnest, start_level, end_level):
    num_spatial_cores = 1
    prev_spatial = True

    # Get outer loop traffic iters from WeightInputBuffer 
    for subnest in list(subnest.values())[start_level:end_level]:
        # Convert everything to a list
        for loop in subnest:
            logger.info("Loop Levels (inner to outer): {0}".format(loop))
            loop_dimension = int(loop['dimension'])
            start = int(loop['start'])
            end = int(loop['end'])
            stride = int(loop['stride'])
            spacetime_dimension = int(loop['spacetime_dimension'])
            size = (end - start)
            if spacetime_dimension != 0:  # If meshX is not 0
                logger.info("\tSpatial Partiton with size {0}".format(size))
                num_spatial_cores *= size
                assert (prev_spatial)  # Make sure spatial loops are the innermost loops
            else:
                prev_spatial = False

    return num_spatial_cores


def get_pe_cycle(subnest, start_level):
    pe_cycle = 1
    # Get inner loop traffic size 
    for subnest in list(subnest.values())[:start_level]:
        for loop in subnest:
            loop_dimension = int(loop['dimension'])
            start = int(loop['start'])
            end = int(loop['end'])
            stride = int(loop['stride'])
            spacetime_dimension = int(loop['spacetime_dimension'])
            size = (end - start)

            # Get the for loop cycles
            if spacetime_dimension == 0:
                pe_cycle *= size
    return pe_cycle


def init_data_size(subnest_info, start_level):
    data_size = {}
    subnest = subnest_info['subnest']
    bufsize = subnest_info['bufsize']

    is_outside_buf = {}
    idx_var_dict = {0: "Weights", 1: "Inputs", 2: "Outputs"}
    for var_name in all_var_names:
        data_size[var_name] = 0
        is_outside_buf[var_name] = False

    # Get inner loop traffic size 
    for i, subnest in enumerate(list(subnest.values())[:start_level]):

        # if the loop is outside the buffer, use the loop bound
        for var_name in idx_var_dict.values():
            if is_outside_buf[var_name]:
                for loop in subnest:
                    loop_dimension = int(loop['dimension'])
                    start = int(loop['start'])
                    end = int(loop['end'])
                    stride = int(loop['stride'])
                    spacetime_dimension = int(loop['spacetime_dimension'])
                    size = (end - start)

                    # Lookup the var affected by the current dim
                    var_arr = dim_var[loop_dimension]
                    if var_name in var_arr:
                        data_size[var_name] *= size
        # else use the timeloop figure
        for var_idx, mem_size in enumerate(list(bufsize.values())[i]):
            if mem_size != 0:
                var_name = idx_var_dict[var_idx]
                data_size[var_name] = mem_size
                is_outside_buf[var_name] = True
    for var_name in idx_var_dict.values():
        logger.info("Data size:\n\t{}={}".format(var_name, data_size[var_name]))
        # print("Data size:\n\t{}={}".format(var_name, schedule['data_size'][var_name]))
    return data_size


def get_iter_space(subnest, start_level, end_level):
    loops = []
    loop_iter = []
    for subnest in list(subnest.values())[start_level:end_level]:
        for loop in subnest:
            item = {}
            item['dimension'] = int(loop['dimension'])
            item['start'] = int(loop['start'])
            item['end'] = int(loop['end'])
            item['stride'] = int(loop['stride'])
            item['spacetime_dimension'] = int(loop['spacetime_dimension'])
            loops.append(item)
            loop_iter.append(item['start'])
    return (loops, loop_iter)


def is_done(loops, loop_iter):
    top_dim = len(loops) - 1
    if loop_iter[top_dim] < loops[top_dim]['end']:
        return False
    else:
        return True


def print_loop_indices(loop_iter):
    loop_indices_str = ', '.join(str(i) for i in loop_iter)
    logger.info("\tLoop Indices: {}".format(loop_indices_str))


def increment(loops, loop_iter):
    loop_carry = True
    for i, loop in enumerate(loops):
        if loop_carry:
            new_iter = loop_iter[i] + loop['stride']
            if new_iter < loop['end']:
                loop_carry = False
                loop_iter[i] = new_iter
                break
            else:
                if i < len(loops) - 1:
                    loop_iter[i] = loop['start']
                else:
                    loop_iter[i] = new_iter


def reset_var_dict_list(var_dict):
    for var in all_var_names:
        var_dict[var] = list()


def reset_var_dict_val(var_dict, val):
    for var in all_var_names:
        var_dict[var] = val


def get_temporal_dim_start(loops):
    # process spatial dims to get send data patterns
    temporal_dim_start = 0
    # loop from inner spatial to outer temporal
    for i, loop in enumerate(loops):
        # if this is spatial
        if loop['spacetime_dimension'] == 0:
            temporal_dim_start = i
            break
    return temporal_dim_start


def get_outer_temp_loopcount(subnest_info, start_level, end_level=None):
    subnest = subnest_info['subnest']
    schedule = {}

    if end_level is None:
        end_level = len(subnest_info['bufsize'])

    # get iteration space starting from PART_LEVEL 
    # loops inside of PART_LEVEL are considered 
    loops, loop_iter = get_iter_space(subnest, start_level, end_level)

    loopcount = 1
    for item in loops:
        # only count time dimension
        if item['spacetime_dimension'] == 0:
            size = (item['end'] - item['start'] + item['stride'] - 1) // item['stride']
            loopcount *= size
    return loopcount


def gen_schedule(subnest_info, start_level, end_level=None):
    logger.info("================ Generate Schedule ================")
    subnest = subnest_info['subnest']
    schedule = {}

    if end_level is None:
        end_level = len(subnest_info['bufsize'])

    schedule['num_spatial_cores'] = get_num_spatial_cores(subnest, start_level, end_level)
    schedule['pe_cycle'] = get_pe_cycle(subnest, start_level)
    data_size = init_data_size(subnest_info, start_level)

    # get iteration space starting from PART_LEVEL 
    # loops inside of PART_LEVEL are considered 
    loops, loop_iter = get_iter_space(subnest, start_level, end_level)

    temporal_dim_start = get_temporal_dim_start(loops)

    spatial_loops = loops[:temporal_dim_start]
    spatial_loop_iter = loop_iter[:temporal_dim_start]
    if not spatial_loops:
        spatial_loops = [{'dimension': start_level, 'start': 0, 'end': 1, 'stride': 1, 'spacetime_dimension': 1}]
        spatial_loop_iter = [0]
    logger.info("Spatial Loops: \n\t{}".format(spatial_loops))
    logger.info("Init Spatial Loops Iter: \n\t{}".format(spatial_loop_iter))

    temporal_loops = loops[temporal_dim_start:]
    temporal_loop_iter = loop_iter[temporal_dim_start:]
    logger.info("Temporal Loops: \n\t{}".format(temporal_loops))
    logger.info("Init Temporal Loops Iter: \n\t{}".format(temporal_loop_iter))
    schedule['temporal_loops'] = temporal_loops
    schedule['temporal_loop_iter'] = copy.deepcopy(temporal_loop_iter)

    # process spatial dims to get send data patterns
    # get the config of the buf data
    # 'Weights' : [True, False] # length = # of spatial level, item = whether the addr varies with the var
    buf_config = {}
    reset_var_dict_list(buf_config)
    for i, loop in enumerate(spatial_loops):
        dimension = loop['dimension']
        for var in all_var_names:
            buf_config[var].append(0)
            if var in dim_var[dimension]:
                buf_config[var][-1] = 1
                # For Inputs, Weights, True indicates multicast, False indicates broadcast
                # For Outputs, True indicates multicast, False indicates reduction
    logger.info("""Spatial Buffer Config (For Inputs and Weights, 1 indicates multicast, 0 indicates broadcast.
For Outputs, True indicates multicast, False indicates reduction.):\n\t{}""".format(buf_config))

    # buf_spatial stores the indices of the spatial buf data
    # 'Weights' : [(0,0), ..., ] # length = # of spatial cores, # item = addr
    buf_spatial = {}
    reset_var_dict_list(buf_spatial)
    while (not is_done(spatial_loops, spatial_loop_iter)):
        # print_loop_indices(spatial_loop_iter)
        for var in all_var_names:
            buf_spatial[var].append(copy.deepcopy(spatial_loop_iter))

        # zero out the irrelevant dims
        for i, loop in enumerate(spatial_loops):
            dimension = loop['dimension']
            for var in all_var_names:
                if not (var in dim_var[dimension]):
                    buf_spatial[var][-1][i] = 0

        increment(spatial_loops, spatial_loop_iter)
    logger.info("Buffer Spatial (Stores the Offsets for the Spatial Buffer Data):\n\t{}".format(buf_spatial))

    # 'Weights' : [True, False] # length = # of spatial level, item = whether the addr varies with the var
    iter_config = {}
    reset_var_dict_list(iter_config)
    for i, loop in enumerate(temporal_loops):
        dimension = loop['dimension']
        start = int(loop['start'])
        end = int(loop['end'])
        stride = int(loop['stride'])
        spacetime_dimension = int(loop['spacetime_dimension'])
        size = (end - start) // stride

        if size > 1:
            for var in all_var_names:
                iter_config[var].append(0)

                if var in dim_var[dimension]:
                    iter_config[var][-1] = 1
                if len(iter_config[var]) > 1:
                    if iter_config[var][-2]:
                        iter_config[var][-1] = 1

    iter_start_dim = {}
    # init all value to # of loop levels
    reset_var_dict_val(iter_start_dim, len(temporal_loops))

    for i, loop in enumerate(temporal_loops):
        for var in all_var_names:
            if iter_start_dim[var] > i:
                dimension = loop['dimension']
                if var in dim_var[dimension]:
                    iter_start_dim[var] = i
    logger.info(
        "Iter Start Dimension (Loop dimension the variable starts vary, inner to outer):\n\t{}".format(iter_start_dim))

    # find the innermost R/S/C loop that is outside output dependent prob K/N/P/Q
    output_start_dim = iter_start_dim['Outputs']
    partial_sum_dep_prob = [0, 1, 4]
    partial_sum_dep_prob_idx = len(temporal_loops)
    for i in range(output_start_dim, len(temporal_loops)):
        loop = temporal_loops[i]
        dimension = loop['dimension']
        if dimension in partial_sum_dep_prob:
            partial_sum_dep_prob_idx = i
            break
    logger.info("Outputs Start Dimension {}, Partial Sum Depend Dimension:{}".format(output_start_dim,
                                                                                     partial_sum_dep_prob_idx))

    #
    steps, iters = new_generate_temp(temporal_loops, iter_start_dim, partial_sum_dep_prob_idx)
    num_steps = len(steps['Inputs'])

    #    for key in steps.keys():
    #        #print(key)
    #        if steps[key] != dup_steps[key]:
    #            print(steps[key])
    #            print(dup_steps[key])

    cost = get_cost(iters, data_size, buf_spatial)

    logger.info("Total Number of Steps: {}".format(num_steps))

    schedule['cost'] = cost
    schedule['data_size'] = data_size
    schedule['buf_config'] = buf_config
    schedule['buf_spatial'] = buf_spatial
    schedule['iter_config'] = iter_config
    schedule['iter_start_dim'] = iter_start_dim
    schedule['steps'] = steps
    schedule['iters'] = iters
    schedule['num_steps'] = num_steps

    logger.info("Total Number of Spatial Cores: {}".format(schedule['num_spatial_cores']))
    logger.info("Total Number of PE Cycles: {}".format(schedule['pe_cycle']))

    return schedule


def get_cost(iters, data_size, buf_spatial):
    logger.info("\tGenerate_Cost")
    cost = {}
    total_cost = 0
    for var in ['Weights', 'Inputs', 'Outputs']:
        addr = construct_addrs_dict(buf_spatial[var])
        cost[var] = iters[var] * data_size[var] * var_bits[var] * len(addr)
        cost[var + '_milp'] = data_size[var] * var_bits[var] * len(addr)
        cost[var + '_milp_spatial'] = data_size[var] * var_bits[var]
        logger.info("\t\t{}: iter={}, data_size={}, var_bits={}, unicast_count={}, cost={}".format(var, iters[var],
                                                                                                   data_size[var],
                                                                                                   var_bits[var],
                                                                                                   len(addr),
                                                                                                   cost[var]))
        total_cost += cost[var]

    var = 'Outputs_Store'
    addr = construct_addrs_dict(buf_spatial['Outputs'])
    cost[var] = iters[var] * data_size['Outputs'] * var_bits['Outputs'] * len(addr)
    cost[var + '_milp'] = data_size['Outputs'] * var_bits['Outputs'] * len(addr)
    logger.info("\t {}: iter={}, data_size={}, var_bits={}, unicast_count={}, cost={}".format(var, iters[var],
                                                                                              data_size['Outputs'],
                                                                                              var_bits['Outputs'],
                                                                                              len(addr), cost[var]))
    total_cost += cost[var]
    cost['Total'] = total_cost
    return cost


def new_generate_temp(temporal_loops, iter_start_dim, partial_sum_dep_prob_idx):
    loop_bound_tup = tuple([int(loop['end']) for loop in temporal_loops])
    loop_level = len(temporal_loops)

    iters = {}
    steps = {}
    for var in ['Weights', 'Inputs', 'Outputs', 'Outputs_Store']:
        steps[var] = np.zeros(loop_bound_tup, dtype=np.uint8)

    for var in ['Weights', 'Inputs']:
        start_dim = iter_start_dim[var]
        np_idx = []

        for dim in range(loop_level):
            if dim < start_dim:
                np_idx.append(0)
            else:
                np_idx.append(slice(None))
        np_idx_tup = tuple(np_idx)
        steps[var][np_idx_tup] = 1
        iters[var] = np.sum(steps[var])

    var = 'Outputs'
    if partial_sum_dep_prob_idx < loop_level:
        start_dim = iter_start_dim[var]
        np_idx = []
        for dim in range(loop_level):
            if dim < start_dim:
                np_idx.append(0)
            elif dim == partial_sum_dep_prob_idx:
                np_idx.append(slice(1, None))
            else:
                np_idx.append(slice(None))
        np_idx_tup = tuple(np_idx)
        steps[var][np_idx_tup] = 1
    iters[var] = np.sum(steps[var])

    var = 'Outputs_Store'
    start_dim = iter_start_dim['Outputs']
    np_idx = []
    for dim in range(loop_level):
        if dim < start_dim:
            np_idx.append(-1)
        else:
            np_idx.append(slice(None))
    np_idx_tup = tuple(np_idx)
    steps[var][np_idx_tup] = 1
    iters[var] = np.sum(steps[var])

    for var in ['Weights', 'Inputs', 'Outputs', 'Outputs_Store']:
        steps[var] = list(steps[var].transpose().flatten())
    # temporal_loop_iter = np.unravel_index(idx, loop_bound)
    return (steps, iters)


if __name__ == "__main__":
    module_name = os.path.basename(__file__).replace(".py", "")
    utils.setup_logging(module_name, logger)
    xml_file = "timeloop-model.map+stats.xml"
    subnest_info = get_subnest_info(xml_file);
    print(subnest_info)
    schedule = gen_schedule(subnest_info, start_level=4, end_level=5)



import logging
import pathlib
import shutil

import cosa.utils
from cosa.parse_workload import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)  # capture everything


def run_config(mapspace, spatial_config, perm_config, factor_config, status_dict=dict(), run_gen_map=True,
               run_gen_tc=False, run_sim_test=False, output_path='output_dir', spatial_configs=None, exe='sim_test',
               valid_check=False, nb_sim=False, outer_loopcount_limit=256, delete_invalid=False):
    mapspace.reset_mapspace(spatial_config, spatial_configs)
    mapspace.update_mapspace(perm_config, factor_config)

    # if it does not pass the check skip it
    if valid_check:
        valid_mapping, _, _, _, _ = mapspace.valid_check()
        if not valid_mapping:
            status_dict = {'run_status': [0]}
            return status_dict
            # raise ValueError()

    # get the config_space_str as key to the status_dict
    # status_dict_key = mapspace.config_space_str(spatial_config, perm_config, factor_config)
    key_strs = [mapspace.arch.config_str(), mapspace.prob.config_str(), mapspace.config_str()[0],
                mapspace.config_str()[1]]

    status_dict_key = '+'.join(key_strs)
    status_dict_val = status_dict.get(status_dict_key)

    if status_dict_val:
        finish_run = True
        if run_gen_map:
            if status_dict_val['run_status'][0] == -1:
                finish_run = False
        if run_gen_tc:
            if status_dict_val['run_status'][1] == -1:
                finish_run = False
        if run_sim_test:
            if status_dict_val['run_status'][2] == -1:
                finish_run = False
        if finish_run:
            logging.info("status_dict: {}".format(status_dict[status_dict_key]))
            # return
            return status_dict_val
    else:
        idx = 0
        # idx can go out of index
        # idx = mapspace.get_idx_from_factor_config(factor_config)
        status_dict_val = {}
        status_dict_val['run_status'] = [-1, -1, -1]
        status_dict_val['cycle_results'] = [-1] * 6  # un-initialized
        status_dict_val['utilized_capacity'] = []
        status_dict[status_dict_key] = status_dict_val

    mapping = mapspace.generate_mapping()

    output_base = pathlib.Path(output_path).resolve()
    output_dir = output_base / mapspace.arch.config_str() / mapspace.prob.config_str() / mapspace.config_str()[0] / \
                 mapspace.config_str()[1]
    # print(output_dir)

    status_dict[status_dict_key]['output_dir'] = str(output_dir)

    output_dir.mkdir(parents=True, exist_ok=True)
    prefix = 'tc'

    map_path = output_dir / 'map_16.yaml'
    logging.info("map_path: {}".format(map_path))
    xml_file = output_dir / 'timeloop-model.map+stats.xml'
    csv_file = output_dir / "{}.csv".format(prefix)
    json_file = output_dir / "{}.json".format(prefix)
    stats_file = output_dir / "{}.summary.json".format(prefix)
    status_dict_file = output_dir / "{}.dict.json".format(prefix)

    # logging.debug("status_dict_before: {}".format(status_dict[status_dict_key]))
    # generate map 
    if run_gen_map:
        # print('run_timeloop> timeloop-model {} {} {}'.format(mapspace.arch.path, mapspace.prob.path, map_path))
        if map_path.exists() and not xml_file.exists():
            status_dict[status_dict_key]['run_status'][0] = 0
        elif not xml_file.exists():
            # logging.info("Run Generate Mapping")
            utils.store_yaml(map_path, mapping)
            success = utils.run_timeloop(mapspace.arch.path, mapspace.prob.path, map_path, cwd=output_dir)

            # if passes timeloop check
            if not success:
                status_dict[status_dict_key]['run_status'][0] = 0
                logging.info("\tInvalid Mapping Detected!")
                if delete_invalid:
                    shutil.rmtree(
                        output_base / mapspace.arch.config_str() / mapspace.prob.config_str() / mapspace.config_str()[
                            0])
                return status_dict[status_dict_key]
            else:
                # logging.info("\tValid Mapping Detected!")
                assert (xml_file.exists())
                status_dict[status_dict_key]['run_status'][0] = 1
        else:
            status_dict[status_dict_key]['run_status'][0] = 1

        # Parse buffer util
        if xml_file.exists():
            subnest_info = get_subnest_info(xml_file)
            bufsize = subnest_info['bufsize']
            pe_cycle = subnest_info['pe_cycle']
            pe_energy = subnest_info['pe_energy']
            cycle = subnest_info['cycle']
            energy = subnest_info['energy']
            status_dict[status_dict_key]['pe_cycle'] = pe_cycle
            status_dict[status_dict_key]['pe_energy'] = pe_energy
            status_dict[status_dict_key]['energy'] = energy
            status_dict[status_dict_key]['cycle'] = cycle
            for buf_idx, (buf_name, buf) in enumerate(bufsize.items()):
                utilized_capacity = 0
                for mem_util in buf:
                    utilized_capacity += mem_util
                status_dict[status_dict_key]['utilized_capacity'].append(utilized_capacity)
    logging.info("Status: {}".format(status_dict[status_dict_key]))
    utils.store_json(status_dict_file, status_dict, indent=4)
    return status_dict[status_dict_key]


def get_perm_size(mapspace, spatial_config, perm_config, factor_config, status_dict, output_path='output_dir',
                  spatial_configs=None, exe='sim_test', valid_check=False, outer_loopcount_limit=None):
    status_dict = run_config(mapspace, spatial_config, perm_config, factor_config, status_dict, run_gen_map=True,
                             run_gen_tc=True, run_sim_test=False, output_path=output_path,
                             spatial_configs=spatial_configs, exe=exe, valid_check=valid_check,
                             outer_loopcount_limit=outer_loopcount_limit)
    sizes = []
    for var in ['Weights', 'Inputs', 'Outputs']:
        sizes.append(status_dict['cost']['{}_milp'.format(var)])
    return sizes


def get_spatial_size(mapspace, spatial_config, perm_config, factor_config, status_dict, output_path='output_dir',
                     spatial_configs=None, exe='sim_test', valid_check=False, outer_loopcount_limit=None):
    status_dict = run_config(mapspace, spatial_config, perm_config, factor_config, status_dict, run_gen_map=True,
                             run_gen_tc=True, run_sim_test=False, output_path=output_path,
                             spatial_configs=spatial_configs, exe=exe, valid_check=valid_check,
                             outer_loopcount_limit=outer_loopcount_limit)
    sizes = []
    for var in ['Weights', 'Inputs', 'Outputs']:
        sizes.append(status_dict['cost']['{}_milp_spatial'.format(var)])
    return sizes



#!/usr/bin/env python3 
import argparse
import logging
import os
import pathlib
import time

import numpy as np
import cosa.run_config as run_config
from cosa.cosa_constants import _A, _B
from cosa.cosa_input_objs import Prob, Arch, Mapspace
from gurobipy import *

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)  # capture everything
logger.disabled = True

try:
    _COSA_DIR = os.path.expanduser(os.environ['COSA_DIR'])
except KeyError:
    _COSA_DIR = os.path.abspath(__file__ + "/../")


def construct_argparser():
    parser = argparse.ArgumentParser(description='Run Configuration')
    parser.add_argument('-o',
                        '--output_dir',
                        type=str,
                        help='Output Folder',
                        default='output_dir',
                        )
    parser.add_argument('-ap',
                        '--arch_path',
                        type=str,
                        help='Hardware Architecture Path',
                        default=f'{_COSA_DIR}/configs/arch/simba.yaml',
                        )
    parser.add_argument('-mp',
                        '--mapspace_path',
                        type=str,
                        help='Mapspace Path',
                        default=f'{_COSA_DIR}/configs/mapspace/mapspace.yaml',
                        )
    parser.add_argument('-pp',
                        '--prob_path',
                        type=str,
                        help='Problem Dimension Path',
                        default=f'{_COSA_DIR}/configs/workloads/resnet50_graph/_outputs_input.2.yaml',
                        )
    return parser


def cosa(prob, arch, A, B, part_ratios, global_buf_idx, Z=None):
    """Run CoSA to generate a mapping with tiling, temporal/spatial, and permutation determined. 
        We currently assume there is a global buffer and only perform the permutation optimization
        at the global buffer level. Will add perm to all level in future version. 

    Args:
        prob: An object defines the layer dimension.
        arch: An object defines the hardware architecture dimension. 
        A: A 2d binary constant matrix that encodes the layer dimension to data tensor relationship.
            1 means related, 0 means unrelated
            Note that the R,S to the input tensor relation is specially handled in the formulation,
            and are specified to 0. 
        B: A 2d binary constant matrix that encodes the data tensor to memory level mapping. 
            It can be derived from the mapspace bypass pattern in Timeloop. 
            Note it is intended to be used for even mapping among different data tensors to different memory levels.
        part_ratios: A 2d array to represent the partition ratios of different data tensors 
            in different memory buffers. 
        global_buf_idx: An index point to the global buffer. 
        Z: Similar to B, but intended for uneven mapping among different data tensors to different memory levels.
            It is a 3d binary constant matrix that encodes the data tensor to memory level mapping.

    Returns: 
        factor_config: A 2d array specifying the allocation decision for each prime factor.
        spatial_config: A 2d array specifying the temporal/spatial decisions for each prime factor.
        perm_config: A 2d array specifyng the ordering of R,S,P,Q,C,K,N factors at each level.  
        run_time: Time-to-solution of CoSA.
    """
    # prime factors 
    prime_factors = prob.prob_factors
    strides = [prob.prob['Wstride'], prob.prob['Hstride']]

    if Z is None:
        Z = []
        for var in _B:
            Z_var = []
            for i, val in enumerate(var):
                rank_arr = [0] * len(var)
                if val == 1:
                    for j in range(i + 1):
                        rank_arr[j] = 1
                Z_var.append(rank_arr)
            Z.append(Z_var)

    factor_config, spatial_config, outer_perm_config, run_time = mip_solver(prime_factors, strides, arch, part_ratios,
                                                                            global_buf_idx=4, A=_A, Z=Z,
                                                                            compute_factor=10, util_factor=-0.1,
                                                                            traffic_factor=1)
    return factor_config, spatial_config, outer_perm_config, run_time


def mip_solver(f, strides, arch, part_ratios, global_buf_idx, A, Z, compute_factor=10, util_factor=-1,
               traffic_factor=1):
    """CoSA mixed integer programming(MIP) formulation."""

    logging.info(f"LAYER {f}")

    num_vars = len(A[0])
    num_mems = len(Z[0])

    m = Model("mip")
    cost = []
    constraints = []

    org = ['spatial', 'temporal']

    M = []

    # ignore DRAM cap
    for i in range(num_mems - 1):
        mem_cap = arch.mem_entries[i]
        mem_cap_arr = []
        for j in range(num_vars):
            var_mem_cap = mem_cap * part_ratios[i][j]
            mem_cap_arr.append(var_mem_cap)
        M.append(mem_cap_arr)

    # log friendly M
    M_log = []
    for i, mem in enumerate(M):
        M_v = []
        for bound in mem:
            if bound == 0:
                # turn 0 to 1 for taking the log
                bound = 1
            M_v.append(bound)
        M_log.append(M_v)

    # spatial constraints
    S = arch.S

    # set the levels to be equal to the number of factors + 4 memory levels 
    perm_levels = 0
    for j, f_j in enumerate(f):
        perm_levels += len(f_j)
    gb_start_level = global_buf_idx

    total_levels = num_mems - 1 + perm_levels
    logging.info(f"total {total_levels} levels")

    x = {}  # x_jn_jn
    for i in range(total_levels):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    x[(i, j, n, k)] = m.addVar(vtype=GRB.BINARY, name=name)
                # sum for each sub factor spatial and temp must be less than 1 
                # NOT equals to one
                spatial_temp_sum = 0
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    spatial_temp_sum += x[(i, j, n, k)]
                m.addConstr(spatial_temp_sum <= 1, "spatial_temp_sum_{}_{}_{}".format(i, j, n))

    # j, n is the loop level 
    # each mapper must have a mapping
    i = 0
    x_row_sums = []
    x_col_sums = []
    # for i in range(total_levels):
    for i in range(gb_start_level, gb_start_level + perm_levels):
        row_sum = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    row_sum += x[(i, j, n, k)]
        m.addConstr(row_sum <= 1, "row_sum_{}".format(i))
        x_row_sums.append(row_sum)

    for j, f_j in enumerate(f):
        for n, f_jn in enumerate(f_j):
            col_sum = 0
            for i in range(total_levels):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    col_sum += x[(i, j, n, k)]
            # assume perm can be interleaved in diff perm level
            m.addConstr(col_sum == 1, "col_sum_{}_{}".format(j, n))
            x_col_sums.append(col_sum)

            # make sure v is one for all outer loop level, once a correlation exists
    # add another relation var v - f, 3 - 7*n loop-level
    s = {}
    y = {}
    for v in range(num_vars):
        for i in range(gb_start_level, gb_start_level + perm_levels):
            row_sum = 0
            y[(v, i)] = m.addVar(lb=0, ub=1, vtype=GRB.INTEGER, name="y({},{})".format(v, i))
            for j, f_j in enumerate(f):
                for n, f_jn in enumerate(f_j):
                    row_sum += x[(i, j, n, 1)] * A[j][v]
            if i > gb_start_level:
                m.addConstr(y[(v, i)] >= y[(v, i - 1)], "y_v_i_sv_{}_{}".format(v, i))
                # can be ==
                m.addConstr(y[(v, i)] >= row_sum, "y_v_i_row_sum_{}_{}".format(v, i))
            else:
                # can be ==
                m.addConstr(y[(v, i)] == row_sum, "y_v_i_row_sum_{}_{}".format(v, i))
            s[(v, i)] = row_sum

    ## exhausively list all scenarios where p or q is inside current mem
    zz = {}
    prefix = 0
    for var in [2, 3]:
        for mem_level in [3]:
            zz[(var, mem_level)] = m.addVar(lb=0, ub=1, vtype=GRB.INTEGER,
                                            name="zz({},{},{})".format(prefix, var, mem_level))
            x_sums = 0
            for n, prime_factor in enumerate(f[var]):
                for inner_mem_level_i in range(mem_level + 1):
                    for k in range(2):
                        filter_in = x[(inner_mem_level_i, var, n, k)]
                        m.addConstr(zz[(var, mem_level)] >= filter_in,
                                    "zz_x_sum_{}_{}_{}_{}_{}_{}".format(prefix, var, n, mem_level, inner_mem_level_i,
                                                                        k))
                        x_sums += filter_in
            m.addConstr(zz[(var, mem_level)] <= x_sums, "z_x_sum_{}_{}_{}".format(prefix, var, mem_level))

    l = {}
    for v in range(num_vars):
        for i in range(gb_start_level, gb_start_level + perm_levels):
            row_sum = 0
            for j, f_j in enumerate(f):
                for n, f_jn in enumerate(f_j):
                    row_sum += np.log2(f[j][n]) * (x[(i, j, n, 1)])
            l[(v, i)] = row_sum

    # Add spatial constraints
    spatial_tile = 0
    for i in range(gb_start_level, gb_start_level + perm_levels):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                spatial_tile += np.log2(f[j][n]) * x[(i, j, n, 0)]
    m.addConstr(spatial_tile <= np.log2(S[gb_start_level]), "spatial_tile_gb_{}".format(prefix))

    for i in range(gb_start_level):
        spatial_tile = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                spatial_tile += np.log2(f[j][n]) * x[(i, j, n, 0)]
        m.addConstr(spatial_tile <= np.log2(S[i]), f"spatial_tile_{prefix}_{i}")

    for i in range(gb_start_level + perm_levels, total_levels):
        spatial_tile = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                spatial_tile += np.log2(f[j][n]) * x[(i, j, n, 0)]
        m.addConstr(spatial_tile <= np.log2(S[i - perm_levels + 1]), f"spatial_tile_{i - perm_levels + 1}")

    # Add inner gb buffer constraints
    buf_util = {}
    for v in range(num_vars):
        for i in range(num_mems):
            buf_util[(i, v)] = 0

    for v in range(num_vars):
        for i_ in range(gb_start_level + perm_levels):
            for i in range(num_mems):
                for j, f_j in enumerate(f):
                    for n, f_jn in enumerate(f_j):
                        factor = 1
                        if v == 1 and j == 2:
                            factor = strides[0]
                        if v == 1 and j == 3:
                            factor = strides[1]

                        if i_ > gb_start_level and i_ < gb_start_level + perm_levels:
                            Z_const = Z[v][i][gb_start_level]
                        else:
                            Z_const = Z[v][i][i_]
                        buf_util[(i, v)] += np.log2(factor * f[j][n]) * (x[(i_, j, n, 0)] + x[i_, j, n, 1]) * A[j][
                            v] * Z_const  # use the i for the cur mem for relationship 
                        # only add once
                        if i == 3 and j in [0, 1] and v == 1:
                            buf_util[(i, v)] += (x[(i_, j, n, 0)] + x[(i_, j, n, 1)]) * (1 - zz[(j + 2, i)]) * np.log2(
                                f[j][n])
                            buf_util[(i, v)] += (x[(i_, j, n, 0)] + x[(i_, j, n, 1)]) * zz[(j + 2, i)] * np.log2(2)

    for v in range(num_vars):
        # excluding DRAM
        for i in range(num_mems - 1):
            if M_log[i][v] > 0:
                m.addConstr(buf_util[(i, v)] <= np.log2(M_log[i][v]), f"buffer_size_{i}_{v}")

                # get compute cost 
    inner_gb_cycles = 0
    for i in range(gb_start_level):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                inner_gb_cycles += np.log2(f[j][n]) * (x[(i, j, n, 1)])

    gb_cycles = 0
    for i in range(gb_start_level, gb_start_level + perm_levels):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                gb_cycles += np.log2(f[j][n]) * (x[(i, j, n, 1)])

    dram_cycles = 0
    for i in range(gb_start_level + perm_levels, total_levels):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                dram_cycles += np.log2(f[j][n]) * (x[(i, j, n, 1)])
    total_compute = inner_gb_cycles + gb_cycles + dram_cycles
    gb_compute = inner_gb_cycles + gb_cycles

    # get traffic cost
    spatial_cost = {}
    for v in range(num_vars):
        size = 0
        for i in range(gb_start_level, gb_start_level + perm_levels):
            for j, f_j in enumerate(f):
                for n, f_jn in enumerate(f_j):
                    size += np.log2(f[j][n]) * (x[(i, j, n, 0)])
        spatial_cost[v] = size

    data_size = {}
    for v in range(num_vars):
        size = 0
        for i in range(gb_start_level):
            for j, f_j in enumerate(f):
                for n, f_jn in enumerate(f_j):
                    # TRICK prioritize spatial
                    factors = 0.8 + 0.04 * i
                    size += factors * np.log2(f[j][n]) * (x[(i, j, n, 0)] + x[i, j, n, 1]) * A[j][v]
        data_size[v] = size

    gb_traffic = {}
    for v in range(num_vars):
        size = 0
        for i in range(gb_start_level, gb_start_level + perm_levels):
            size += l[(v, i)] * y[(v, i)]
        gb_traffic[v] = size

        # use the last level gb y for DRAM 
    dram_traffic = {}
    for v in range(num_vars):
        corr = y[(v, gb_start_level + perm_levels - 1)]
        i = gb_start_level + perm_levels  # DRAM 
        size = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                size += np.log2(f[j][n]) * (x[(i, j, n, 1)])  # * corr 
        dram_traffic[v] = size

    total_util = 0
    for i in range(gb_start_level):
        # for each memory and each variable there is a constraint
        for v in range(num_vars):
            # make weight util more important since it directly comes from dram
            factor = 1.01 if i == 2 else 1
            total_util += buf_util[(i, v)] * factor

    total_traffic = 0
    for v in range(num_vars):
        #  TRICKS
        if v == 0:
            # encode dram latency for weights
            factor = 1.01
        else:
            factor = 1

        total_traffic += 0.99 * data_size[v] + 0.99 * spatial_cost[v] + gb_traffic[v] + dram_traffic[v] * factor

    # ========================== user-defined objective function ========================== #
    cosa_obj = total_util * util_factor + total_compute * compute_factor + total_traffic * traffic_factor

    max_it = m.addVar(vtype=GRB.CONTINUOUS, name="max_it")
    its = []
    its.append(m.addVar(vtype=GRB.CONTINUOUS, name="a"))
    m.addConstr(its[-1] == total_traffic, "total_traffic")
    its.append(m.addVar(vtype=GRB.CONTINUOUS, name="b"))
    m.addConstr(its[-1] == total_compute, "total_compute")
    m.addConstr(max_it == max_(its), name="max_it_constr")

    total_util_var = m.addVar(vtype=GRB.CONTINUOUS, name="total_util_var")
    total_comp_var = m.addVar(vtype=GRB.CONTINUOUS, name="total_comp_var")
    total_traf_var = m.addVar(vtype=GRB.CONTINUOUS, name="total_traf_var")

    # cycle count = total max 3 * all log factors variables 
    m.addConstr(total_util_var == total_util, "total_util_constraint")
    m.addConstr(total_comp_var == total_compute, "total_comp_constraint")
    m.addConstr(total_traf_var == total_traffic, "total_traf_constraint")

    m.ModelSense = GRB.MINIMIZE
    m.setObjective(cosa_obj, GRB.MINIMIZE)

    # optimize for the objective function
    milp_time = 0
    begin_time = time.time()
    m.optimize()
    end_time = time.time()
    milp_runtime = end_time - begin_time

    # output all constraints and variables
    m.write("debug.lp")

    result_dict = {}
    for variable in m.getVars():
        # logging.debug("Variable %s: Value %s" % (variable.varName, variable.x))
        assert (variable.varName not in result_dict)
        result_dict[variable.varName] = variable.x
    logging.debug('Obj: %g' % m.objVal)

    all_x = np.zeros((total_levels, perm_levels, 2))
    for i in range(total_levels):
        level_idx = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    val = result_dict[name]
                    all_x[i, level_idx, k] = val
                level_idx += 1
    np.set_printoptions(precision=0, suppress=True)

    var_outer_perm_config = [-1] * perm_levels
    outer_perm_config = [-1] * perm_levels
    x_arr = np.zeros((perm_levels, perm_levels, 2))
    for i in range(gb_start_level, gb_start_level + perm_levels):
        level_idx = 0
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    val = result_dict[name]
                    x_arr[i - gb_start_level, level_idx, k] = val
                name = "X({},{},{},{})".format(i, j, n, 1)
                val = result_dict[name]
                if val == 1:
                    var_outer_perm_config[i - gb_start_level] = j
                level_idx += 1
    logging.info(f'var_outer_perm_config: {var_outer_perm_config}')

    y_arr = np.zeros((num_vars, perm_levels))
    for v in range(num_vars):
        for i in range(gb_start_level, gb_start_level + perm_levels):
            row_sum = 0
            val = result_dict["y({},{})".format(v, i)]
            y_arr[v, i - gb_start_level] = val

    # Merge the permutation, taking the first appearance of a prob to be the
    merge_outer_perm_config = []
    for i, var in enumerate(var_outer_perm_config):
        if var != -1 and var not in merge_outer_perm_config:
            merge_outer_perm_config.append(var)

    for i in range(len(f)):
        if i not in merge_outer_perm_config:
            merge_outer_perm_config.append(i)

    logging.info("var idx as the value {}".format(var_outer_perm_config))
    logging.info("merged var idx as the value {}".format(merge_outer_perm_config))

    outer_perm_config = [1] * len(f)
    for i, var in enumerate(merge_outer_perm_config):
        outer_perm_config[var] = i

    logging.info("ordering idx as the value {}".format(outer_perm_config))

    # init factor_config 
    # DRAM is the last level
    factor_config = []
    spatial_config = []
    dram_level = -1
    for j, f_j in enumerate(f):
        sub_factor_config = []
        sub_spatial_config = []
        for n, f_jn in enumerate(f_j):
            sub_factor_config.append(dram_level)
            sub_spatial_config.append(0)
        factor_config.append(sub_factor_config)
        spatial_config.append(sub_spatial_config)

    for i in range(gb_start_level):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                if f[j][n] == 1:
                    factor_config[j][n] = num_mems - 1
                    spatial_config[j][n] = 0
                    continue
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    val = result_dict[name]
                    if val >= 0.9:
                        factor_config[j][n] = i
                        if k == 0:
                            spatial_config[j][n] = 1

    for i in range(gb_start_level + perm_levels, total_levels):
        for j, f_j in enumerate(f):
            for n, f_jn in enumerate(f_j):
                if f[j][n] == 1:
                    factor_config[j][n] = num_mems - 1
                    spatial_config[j][n] = 0
                    continue

                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    val = result_dict[name]
                    if val >= 0.9:
                        if k == 0:
                            raise ValueError('Invalid Mapping')
                        factor_config[j][n] = i - perm_levels + 1

    # set to -1 for not specified 
    for j, f_j in enumerate(f):
        for n, f_jn in enumerate(f_j):
            for i in range(gb_start_level, gb_start_level + perm_levels):
                for k in range(2):
                    name = "X({},{},{},{})".format(i, j, n, k)
                    val = result_dict[name]
                    if val >= 0.9:
                        factor_config[j][n] = gb_start_level
                        if k == 0:
                            spatial_config[j][n] = 1

    logging.info(f"prime factors: {f}")
    logging.info(f"factor configs: {factor_config}")
    logging.info(f"spatial configs: {spatial_config}")

    return (factor_config, spatial_config, outer_perm_config, milp_runtime)


def run_timeloop(prob_path, arch_path, mapspace_path, output_path):
    # init
    status_dict = {}
    prob = Prob(prob_path)
    arch = Arch(arch_path)

    # An object defines the user-defined bypass pattern. 
    mapspace = Mapspace(mapspace_path)
    mapspace.init(prob, arch)

    # even mapping
    B = _B
    Z = None

    # uneven mapping config
    # Z = _Z
    # B = None
 
    # partition ratios for W, IA, OA
    part_ratios = [
        [1, 0, 0],
        [0, 0, 1],
        [1, 0, 0],
        [0, 1, 0],
        [0, 0.25, 0.75],
        [0.33, 0.33, 0.33],
    ]
    factor_config, spatial_config, outer_perm_config, run_time = cosa(prob, arch, _A, B, part_ratios, global_buf_idx=4,
                                                                      Z=Z)

    update_factor_config = factor_config
    spatial_to_factor_map = {}
    idx = arch.mem_levels
    for i, val in enumerate(arch.S):
        if val > 1:
            spatial_to_factor_map[i] = idx
            idx += 1
    logging.info(f'spatial_to_factor_map: {spatial_to_factor_map}')

    for j, f_j in enumerate(prob.prob_factors):
        for n, f_jn in enumerate(f_j):
            # if is mapped to spatial, look up the combined index
            if spatial_config[j][n] == 1:
                idx = factor_config[j][n]
                update_factor_config[j][n] = spatial_to_factor_map[idx]

    logging.info(f'update_factor_config: {update_factor_config}')
    perm_config = mapspace.get_default_perm()
    perm_config[4] = outer_perm_config

    status_dict = {}
    try:
        spatial_configs = []
        results = run_config.run_config(mapspace, None, perm_config, update_factor_config, status_dict,
                                        run_gen_map=True, run_gen_tc=False, run_sim_test=False, output_path=output_path,
                                        spatial_configs=spatial_configs, valid_check=False, outer_loopcount_limit=100)
        logging.info(f'status_dict: {status_dict}')
    except:
        logging.error('Error: invalid schedule.')

    return status_dict

def run_cosa():
    parser = construct_argparser()
    args = parser.parse_args()

    prob_path = pathlib.Path(args.prob_path).resolve()
    arch_path = pathlib.Path(args.arch_path).resolve()
    mapspace_path = pathlib.Path(args.mapspace_path).resolve()
    output_path = args.output_dir

    run_timeloop(prob_path, arch_path, mapspace_path, output_path)


if __name__ == "__main__":
    run_cosa()


import collections
import errno
import json
import logging
import math
import os
import pathlib
import pickle
import random
import re
import string
import subprocess
import sys
from time import strftime, gmtime

import numpy as np
import yaml

logger = logging.getLogger(__name__)
logger.setLevel(logging.NOTSET)  # capture everything
logger.disabled = True


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class OrderedDefaultdict(collections.OrderedDict):
    """ A defaultdict with OrderedDict as its base class. """

    def __init__(self, default_factory=None, *args, **kwargs):
        if not (default_factory is None
                or isinstance(default_factory, collections.Callable)):
            raise TypeError('first argument must be callable or None')
        super(OrderedDefaultdict, self).__init__(*args, **kwargs)
        self.default_factory = default_factory  # called by __missing__()

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key, )
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):  # optional, for pickle support
        args = (self.default_factory,) if self.default_factory else tuple()
        return self.__class__, args, None, None, self.iteritems()

    def __repr__(self):  # optional
        return '%s(%r, %r)' % (self.__class__.__name__, self.default_factory,
                               list(self.iteritems()))


def delete_dramsim_log(path=pathlib.Path('.')):
    files = path.glob('dramsim.*.log')
    for f in files:
        f.unlink()


def setup_logging(module_name, logger):
    # logging setup
    def logfilename():
        """ Construct a unique log file name from: date + 16 char random. """
        timeline = strftime("%Y-%m-%d--%H-%M-%S", gmtime())
        randname = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(16))
        return module_name + "-" + timeline + "-" + randname + ".log"

    # log to file
    full_log_filename = logfilename()
    fileHandler = logging.FileHandler(full_log_filename)
    # formatting for log to file
    # TODO: filehandler should be handler 0 (firesim_topology_with_passes expects this
    # to get the filename) - handle this more appropriately later
    logFormatter = logging.Formatter("%(asctime)s [%(funcName)-12.12s] [%(levelname)-5.5s]  %(message)s")
    fileHandler.setFormatter(logFormatter)
    fileHandler.setLevel(logging.NOTSET)  # log everything to file
    logger.addHandler(fileHandler)

    # log to stdout, without special formatting
    consoleHandler = logging.StreamHandler(stream=sys.stdout)
    consoleHandler.setLevel(logging.DEBUG)  # show only INFO and greater in console
    logger.addHandler(consoleHandler)


def dict_append_val(d, k, v):
    if k in d.keys():
        d[k].append(v)
    else:
        d[k] = [v]


def gen_makefile(x_size, y_size):
    makefile_str = """EXT_DRAMSIM2_PATH = ../DRAMSim2
LIBS = -ldramsim

DEBUG_LEVEL ?= 1

USER_FLAGS += -DDISABLE_PACER -DDEBUG_LEVEL=$(DEBUG_LEVEL) -std=c++11 -I$(EXT_DRAMSIM2_PATH) -L$(EXT_DRAMSIM2_PATH) $(LIBS) -DNOC_X={} -DNOC_Y={}
include ../unittests_Makefile
""".format(x_size, y_size)

    with open("Makefile", "w") as f:
        f.write(makefile_str)


def compile_sim():
    # subprocess.check_call(['make', 'clean'],\
    #        cwd=os.getcwd())
    orig_sim_test = pathlib.Path('sim_test')
    if orig_sim_test.exists():
        orig_sim_test.unlink()
    try:
        out = subprocess.check_output(['make'], \
                                      stderr=subprocess.PIPE, cwd=os.getcwd())

        logger.info('compile_sim> make')
    except subprocess.CalledProcessError as grepexc:
        # print "error code", grepexc.returncode, grepexc.output
        if grepexc.returncode == 2:
            pass
            # logger.error(err)


def run_sim(prefix, stdout=None, stderr=None, timeout=None, exe='sim_test', nb=False):
    print("./{} {}".format(exe, prefix))
    try:
        exe = str(exe)
        if nb:
            p = subprocess.check_call(['./' + exe, str(prefix)], \
                                      stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, cwd=os.getcwd(),
                                      timeout=timeout)
        else:
            p = subprocess.check_call(['./' + exe, str(prefix)], \
                                      stdout=stdout, stderr=stderr, cwd=os.getcwd(), timeout=timeout)
        logger.info('run_sim> ./' + exe + ' ' + str(prefix))
        return True
    except:
        return False


def run_timeloop(arch, prob, mapp, cwd=os.getcwd(), stdout=None, stderr=None):
    try:
        p = subprocess.check_call(['timeloop-model', str(arch), str(prob), str(mapp)], \
                                  cwd=cwd, stdout=stdout, stderr=stderr)
        logger.info('run_timeloop> timeloop-model {} {} {}'.format(arch, prob, mapp))
        return True
    except:
        return False


def run_timeloop_mapper(arch, prob, mapspace, cwd=os.getcwd(), stdout=None, stderr=None):
    try:
        arch_path = pathlib.Path('timeloop_configs/arch/simba_large.yaml').resolve()
        mapspace_path = pathlib.Path('timeloop_configs/mapspace/mapspace_io.yaml').resolve()
        p = subprocess.check_call(['timeloop-mapper', arch_path, mapspace_path, str(prob), ], \
                                  cwd=cwd, stdout=stdout, stderr=stderr)
        logger.info('run_timeloop> timeloop-mapper {} {} {}'.format(arch, mapspace, prob))
        return True
    except:
        return False


def run_timeloop_mapper_optimal(arch, prob, mapspace, cwd=os.getcwd(), stdout=None, stderr=None):
    try:
        arch_path = pathlib.Path('timeloop_configs/arch/simba_large.yaml').resolve()
        mapspace_path = pathlib.Path('timeloop_configs/mapspace/mapspace_io_optimal.yaml').resolve()
        p = subprocess.check_call(['timeloop-mapper', arch_path, mapspace_path, str(prob), ], \
                                  cwd=cwd, stdout=stdout, stderr=stderr)
        logger.info('run_timeloop> timeloop-mapper {} {} {}'.format(arch, mapspace, prob))
        return True
    except:
        return False


def run_timeloop_mapper_linear(arch, prob, mapspace, cwd=os.getcwd(), stdout=None, stderr=None):
    try:
        print("Running LINEAR PRUNING")
        arch_path = pathlib.Path('timeloop_configs/arch/simba_large.yaml').resolve()
        mapspace_path = pathlib.Path('timeloop_configs/mapspace/mapspace_io_linear.yaml').resolve()
        p = subprocess.check_call(['timeloop-mapper', arch_path, mapspace_path, str(prob), ], \
                                  cwd=cwd, stdout=stdout, stderr=stderr)
        logger.info('run_timeloop> timeloop-mapper {} {} {}'.format(arch, mapspace, prob))
        return True
    except:
        return False


def run_timeloop_mapper_hybrid(arch, prob, mapspace, cwd=os.getcwd(), stdout=None, stderr=None):
    try:
        print("Running HYBRID Random Index + Linear Pruning")
        arch_path = pathlib.Path('timeloop_configs/arch/simba_large.yaml').resolve()
        mapspace_path = pathlib.Path('timeloop_configs/mapspace/mapspace_io_hybrid.yaml').resolve()
        p = subprocess.check_call(['timeloop-mapper', arch_path, mapspace_path, str(prob), ], \
                                  cwd=cwd, stdout=stdout, stderr=stderr)
        logger.info('run_timeloop> timeloop-mapper {} {} {}'.format(arch, mapspace, prob))
        return True
    except:
        return False


def store_json(json_path, data, indent=None):
    with open(json_path, "w") as f:
        json.dump(data, f, indent=indent)


def parse_json(json_path):
    with open(json_path) as f:
        data = json.load(f)
    return data


def parse_yaml(yaml_path):
    with open(yaml_path, 'r') as f:
        data = yaml.full_load(f)
    return data


def store_yaml(yaml_path, data):
    with open(yaml_path, 'w') as f:
        yaml.dump(data, f)


def parse_pickle(pickle_path):
    with open(pickle_path, 'rb') as f:
        data = pickle.load(f)
    return data


def store_pickle(pickle_path, data):
    with open(pickle_path, 'wb') as f:
        pickle.dump(data, f)


# Return prime values from small to large
def get_prime_factors(value):
    factors = []

    # Print the number of two's that divide n 
    while (value % 2 == 0):
        factors.append(2)
        value = value // 2

    # value must be odd at this point 
    # so a skip of 2 ( i = i + 2) cavalue be used 
    for i in range(3, int(math.sqrt(value)) + 1, 2):

        # while i divides value , print i ad divide value 
        while (value % i == 0):
            factors.append(i)
            value = value // i

            # Condition if value is a prime
    # number greater than 2 
    if value > 2:
        factors.append(value)
    if len(factors) == 0:
        factors.append(1)

    return factors


# remove factor from prime factors
# prob_factors is passed by ref
def update_prime_factors(prob_factors, prob_idx, factor):
    logger.debug("prob_factors {}".format(prob_factors))
    rm_idx = []
    orig_factor = factor
    # print(prob_factors)
    for i, val in enumerate(prob_factors[prob_idx]):
        if factor % val == 0:
            rm_idx.append(i)
            factor = factor // val
        if factor == 1:
            break

    vals = []
    rm_idx.reverse()
    for i in rm_idx:
        vals.append(prob_factors[prob_idx].pop(i))

    if len(prob_factors[prob_idx]) == 0:
        prob_factors[prob_idx] = [1]
    pop_val = 1
    # print(prob_factors)
    logger.debug("updated_prob_factors {}".format(prob_factors))
    for i in vals:
        pop_val = pop_val * i
    assert (pop_val == orig_factor)


# shrink the space
def shrink_factor_space(prob_factors):
    # val get rid of factor of 1
    for prob_idx, _ in enumerate(prob_factors):
        if len(prob_factors[prob_idx]) == 1:
            if prob_factors[prob_idx][0] == 1:
                prob_factors[prob_idx] = []


def compose_prob(file_path, prob_dict):
    input_dict = {}
    input_dict["problem"] = prob_dict
    input_dict['problem']['shape'] = 'cnn-layer'
    input_dict['problem']['Wstride'] = 1
    input_dict['problem']['Hstride'] = 1
    input_dict['problem']['Wdilation'] = 1
    input_dict['problem']['Hdilation'] = 1
    store_yaml(file_path, input_dict)


def get_perm_arr_from_val(val, perm_arr_tup, prob_levels):
    order_idx = np.unravel_index(val, perm_arr_tup)
    perm_idx = list(range(prob_levels))
    perm_arr = [-1] * len(perm_arr_tup)
    # idx is the idx into [0,1,2,3..,6]
    for i, idx in enumerate(order_idx):
        perm_arr[i] = perm_idx[idx]
        del perm_idx[idx]
    return perm_arr


# input perm_arr [0, 3, 5, 6, 2, 4, 1] -> order_idx 
# order_idx -> val 333 
def get_val_from_perm_arr(self, perm_arr, perm_arr_tup, prob_levels):
    assert (len(perm_arr) == len(perm_arr_tup))
    order_idx = [0] * len(perm_arr)
    perm_idx = list(range(prob_levels))
    for i, idx in enumerate(perm_arr):
        for j, val in enumerate(perm_idx):
            if val == idx:
                order_idx[i] = j
                del perm_idx[j]
                break
    val = np.ravel_multi_index(tuple(order_idx), perm_arr_tup)
    return val


def to_config_str_key(configs):
    configs_arr = [str(v) for v in configs]
    configs_str = "_".join(configs_arr)
    return configs_str


def get_correlation(a, b):
    a = np.array(a)
    b = np.array(b)
    cor = np.corrcoef(a, b)[0, 1]
    return cor


def get_cor_stats(status_dicts):
    arr = []
    for d in status_dicts:
        # baseline
        l_arr = [d['cycle_results'][0]]
        l_arr.append(d['cycle_results'][1])
        l_arr.append(d['cycle_results'][2])
        l_arr.append(d['cycle_results'][4])
        l_arr.append(d['cycle_results'][5])
        l_arr.append(d['cost']['Total'])
        l_arr.append(d['hop_cost'][0] + d['hop_cost'][1])
        l_arr.append(d['hop_cost'][0])
        l_arr.append(d['hop_cost'][1])
        l_arr.append((d['utilized_capacity'][1] + d['utilized_capacity'][2] + d['utilized_capacity'][3]))
        l_arr.append((d['utilized_capacity'][1] + d['utilized_capacity'][2] + d['utilized_capacity'][3] +
                      d['utilized_capacity'][4]))

        l_arr.append((d['utilized_capacity'][1] * d['utilized_capacity'][2] * d['utilized_capacity'][3]) ** (1 / 128))
        l_arr.append((d['utilized_capacity'][1] * d['utilized_capacity'][2] * d['utilized_capacity'][3] *
                      d['utilized_capacity'][4]) ** (1 / 128))
        l_arr.append((np.log2(1 + d['utilized_capacity'][1]) + np.log2(1 + d['utilized_capacity'][2]) + np.log2(
            1 + d['utilized_capacity'][3])))
        l_arr.append((np.log2(1 + d['utilized_capacity'][1]) + np.log2(1 + d['utilized_capacity'][2]) + np.log2(
            1 + d['utilized_capacity'][3]) + np.log2(1 + d['utilized_capacity'][4])))

        arr.append(l_arr)
    arr_np = np.array(arr).T

    cors = []
    for i in range(1, len(l_arr)):
        cor = get_correlation(arr_np[0, :], arr_np[i, :])
        cors.append(cor)

    print(arr_np[-3, :])
    x_label = "Total_Cycles"
    print(f"{x_label}-Unicast_Cycles: {cors[0]}")
    print(f"{x_label}-Multicast_Cycles: {cors[1]}")
    print(f"{x_label}-DRAM_READ_Cycles: {cors[2]}")
    print(f"{x_label}-DRAM_WRITE_Cycles: {cors[3]}")
    print(f"{x_label}-Cost: {cors[4]}")
    print(f"{x_label}-Total_Hops: {cors[5]}")
    print(f"{x_label}-Unicast_Hops: {cors[6]}")
    print(f"{x_label}-Multicast_Hops: {cors[7]}")
    print(f"{x_label}-BufSumGBEx: {cors[8]}")
    print(f"{x_label}-BufSumGBIn: {cors[9]}")
    print(f"{x_label}-BufProdGBEx ** (1/128): {cors[10]}")
    print(f"{x_label}-BufProdGBIn ** (1/128): {cors[11]}")
    print(f"{x_label}-BufProdGBEx log): {cors[12]}")
    print(f"{x_label}-BufProdGBIn log: {cors[13]}")

    return cors


def get_invalid_samples(temp_file):
    total_invalid = 0
    with open(temp_file, 'r') as f:
        lines = f.readlines()
    for line in lines:
        m = re.match(r".*MATCHLIB_TOTAL: (.*), MATCHLIB_VALID.*", line)
        if m:
            # print(m)
            invalid = m.group(1)
            total_invalid += int(invalid)
    return total_invalid


def get_runtime(summary_out):
    runtime_dict = {}

    with open(summary_out, 'r') as f:
        lines = f.readlines()
    key = None
    for line in lines:
        m = re.match(r".*simba_(.*)\/(.*)$", line)
        if m:
            key = m.group(2)
        else:
            m = re.match(r"Elasped time for .*find solution with .* is: (.*)$", line)
            if m:
                runtime_dict[key] = float(m.group(1))
    return runtime_dict


def update_arch(arch_dict, mem_entries, out_path):
    for i, mem in enumerate(arch_dict['arch']['storage']):
        # DRAM has no entries
        if i < len(arch_dict['arch']['storage']) - 1:
            arch_dict['arch']['storage'][i]['entries'] = mem_entries[i]
    store_yaml(out_path, arch_dict)



#!/usr/bin/env python3 
import copy
import logging
import pathlib

import numpy as np
import cosa.run_config

import cosa.utils as utils
from cosa.parse_workload import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)  # capture everything


class Prob(object):
    """Problem space with layer dimension, stride and dilation defined.
    
    Attributes: 
        prob: A layer dimemsion dictionary. 
            R, S represent the weight filter width and height.
            P, Q represent the output feature map width and height.
            C represents the input channel size. 
            K represents the output channel size.
            N represents the batch size. 
            Wstride, Hstride represent the width and height dimension stride.
            Wdilation, Hdilation represent the width and height dimension dilation.
        prob_bound:  A 1d array with layer dimension value for R,S,P,Q,C,K,N
            e.g. [1,1,1,2,3,4,5]
        prob_factors:  A 2d array with all prime factors generated from each dimension
            e.g. [[1],[1],[1],[2],[3],[2,2],[5]] 
    """

    def __init__(self, prob_path):
        """Initialize the layer dimension from an input yaml file. 

            Example input yaml file format: 
                problem:
                  C: 3
                  Hdilation: 1
                  Hstride: 2
                  K: 64
                  N: 1
                  P: 112
                  Q: 112
                  R: 7
                  S: 7
                  Wdilation: 1
                  Wstride: 2
                  shape: cnn-layer


        Args: 
            prob_path: Path to the yaml file that defines the convolution layer dimensions. 
        """
        # defines the dimension index for 7 major loop bounds 
        self.prob_idx_name_dict = {0: 'R', 1: 'S', 2: 'P', 3: 'Q', 4: 'C', 5: 'K', 6: 'N'}
        self.prob_name_idx_dict = {v: k for k, v in self.prob_idx_name_dict.items()}

        self.prob_bound = [-1] * len(self.prob_name_idx_dict)
        self.prob_factors = []
        for i in range(len(self.prob_name_idx_dict)):
            self.prob_factors.append([])

        self.prob_levels = len(self.prob_idx_name_dict.items())

        self.path = prob_path.resolve()
        prob_dict = utils.parse_yaml(self.path)
        self.prob = prob_dict['problem']

        for key, value in self.prob.items():
            if ('stride' in key or 'dilation' in key):
                continue
            if (key == 'shape'):
                continue
            prob_idx = self.prob_name_idx_dict[key]
            self.prob_bound[prob_idx] = value
            self.prob_factors[prob_idx] = utils.get_prime_factors(value)

    def config_str(self):
        """Returnsthe key str name for representing a unique layer."""
        val_arr = []
        for value in self.prob_bound:
            val_arr.append(str(value))
        keys = ['Wstride', 'Hstride', 'Wdilation', 'Hdilation']
        val_arr.extend([str(self.prob[key]) for key in keys])
        val_str = "_".join(val_arr)
        return val_str

    def print(self):
        print(self.__dict__)


class Arch(object):
    """ Hardware architecture specifyng number of hardware instances and buffer capacity.
    
    Attributes: 
        mem_instances: number of memory instances per chip.
        mem_entries: number of valid memory entries.
    """

    def __init__(self, arch_path):
        """Initialize the hardware architecture details from an input yaml file. 

        Args: 
            arch_path: Path to the yaml file that defines the hardware architecture constraints. 
        """

        self.path = arch_path.resolve()
        arch_dict = utils.parse_yaml(self.path)

        # arch config version, please add a postfix _v3 to
        # the yaml filename if a new version is used
        version = 'v3' if '_v3' in self.path.name else 'v1'

        # mem instance size for each 
        self.mem_instances = []
        self.mem_entries = []

        # name to idx lookup
        self.mem_idx = {}

        # idx to name lookup
        self.mem_name = {}

        if version == 'v1':
            self.arch = arch_dict['arch']
            for key, value in self.arch.items():
                setattr(self, key, value)
            for i, mem in enumerate(self.storage):
                self.mem_idx[mem['name']] = i
                self.mem_name[i] = mem['name']
                self.mem_instances.append(mem['instances'])
                if i < len(self.storage) - 1:
                    self.mem_entries.append(mem['entries'])
        elif version == 'v3':
            self.dram = arch_dict['architecture']['subtree'][0]['local'][0]
            self.global_buf = arch_dict['architecture']['subtree'][0]['subtree'][0]['local'][0]
            self.pe_buf = arch_dict['architecture']['subtree'][0]['subtree'][0]['subtree'][0]['local']
            idx = 0
            for i, mem in enumerate(self.pe_buf[::-1]):
                if mem['class'] == 'SRAM' or mem['class'] == 'regfile':
                    self.mem_idx[mem['name']] = idx
                    self.mem_name[idx] = mem['name']
                    self.mem_instances.append(mem['attributes']['instances'])
                    self.mem_entries.append(mem['attributes']['entries'])
                    idx += 1

            self.mem_idx[self.global_buf['name']] = idx
            self.mem_name[idx] = self.global_buf['name']
            self.mem_instances.append(self.global_buf['attributes']['instances'])
            self.mem_entries.append(self.global_buf['attributes']['entries'])
            idx += 1
            self.mem_idx[self.dram['name']] = idx
            self.mem_name[idx] = self.dram['name']
            self.mem_instances.append(self.dram['attributes']['instances'])
            self.arch = {"instances": self.mem_instances, "entries": self.mem_entries}
        self.mem_levels = len(self.mem_idx.items())
        self.S = self.gen_spatial_constraint()

    def gen_spatial_constraint(self):
        """Generate spatial constraints."""
        S = []
        inner_instances = self.mem_instances[0]
        for i in self.mem_instances:
            if i != 0:
                S.append(inner_instances // i)
                inner_instances = i
        return S

    def config_str(self):
        """Return the filename for the input yaml with postfix."""
        return self.path.stem

    def print(self):
        print(self.__dict__)


class Mapspace(object):
    """ Mapping Space"""

    def __init__(self, path):
        mapspace_dict = utils.parse_yaml(path)
        self.mapspace = mapspace_dict['mapspace']

        for key, value in self.mapspace.items():
            setattr(self, key, value)

        self.var_idx_dict = {'Weights': 0, 'Inputs': 1, 'Outputs': 2}
        self.var_name_dict = {v: k for k, v in self.var_idx_dict.items()}
        self.org_idx_dict = {'spatial': 0, 'temporal': 1}
        self.org_name_dict = {v: k for k, v in self.org_idx_dict.items()}
        self.config_idx_dict = {'perm': 0, 'factor': 1}
        self.mapspace = None
        self.bypass = None
        self.arch = None
        self.prob = None
        self.factor_space = None
        self.factor_config_tup = None
        self.perm_space = None
        self.spatial_space = None

    def init(self, prob, arch, use_valid_spatial_levels=True):
        self.prob = prob
        self.arch = arch

        # init bypass 
        # order for the spatial dim indicates XY spatial arrangement if X!=Y
        self.bypass = np.zeros((arch.mem_levels, len(self.var_idx_dict.items())), dtype=np.int8)

        # parse bypass info
        for constraint in self.constraints:
            if constraint['type'] == 'datatype':
                mem_idx = arch.mem_idx[constraint['target']]
                bypass_vars = constraint['bypass']
                if bypass_vars is not None:
                    for var in bypass_vars:
                        var_idx = self.var_idx_dict[var]
                        self.bypass[mem_idx][var_idx] = 1

        # init valid spatial level from arch
        self.valid_spatial_levels = self.get_valid_spatial_levels(use_valid_spatial_levels)
        self.valid_spatial_factors = [1] * arch.mem_levels
        for mem_level, valid_spatial_factors in self.valid_spatial_levels:
            self.valid_spatial_factors[mem_level] = valid_spatial_factors

        self.prob_var_dict = {
            0: ["Weights"],
            1: ["Weights"],
            2: ["Inputs", "Outputs"],
            3: ["Inputs", "Outputs"],
            4: ["Inputs", "Weights"],
            5: ["Outputs", "Weights"],
            6: ["Inputs", "Outputs"],
        }

        # parse memory-related variables 
        self.var_mem_dict = self.get_var_mem()

        # init
        spatial_configs = [1] * self.prob.prob_levels
        self._reset_mapspace(spatial_configs)

        # init the ordering
        # inner to outer 
        # idx of the prob dimension
        perm_config = self.get_default_perm()

        # set up permutation space 
        self.perm_space = np.math.factorial(self.prob.prob_levels) * np.ones(
            (arch.mem_levels, len(self.org_idx_dict.items())))
        self.perm_space_tup = tuple(self.perm_space.flatten())

        self.perm_arr_tup = tuple(range(self.prob.prob_levels, 0, -1))

        # 7 x prob_factors
        factor_config = self.get_default_factor()
        self.update_mapspace(perm_config, factor_config)

    # input 333 -> order_idx [0, 2, 3, 3, 1, 1, 0]
    # order_idx -> perm_arr [0, 3, 5, 6, 2, 4, 1]
    def get_perm_arr_from_val(self, val):
        perm_arr = utils.get_perm_arr_from_val(val, self.perm_arr_tup, self.prob.prob_levels)
        return perm_arr

    # input perm_arr [0, 3, 5, 6, 2, 4, 1] -> order_idx 
    # order_idx -> val 333 
    def get_val_from_perm_arr(self, perm_arr):
        val = utils.get_val_from_perm_arr(self, perm_arr, self.perm_arr_tup, self.prob.prob_levels)
        return val

    def get_perm_config_from_idx(self, idx):
        perm_idx_arr = np.unravel_index(idx, self.perm_config_tup)
        perm_idx = perm_idx_arr.reshape(self.perm_config)
        return perm_idx

    def get_idx_from_perm_config(self, perm_config):
        perm_config_tup = tuple(perm_config.flatten())
        idx = np.ravel_multi_index(perm_config_tup, self.perm_space_tup)
        return idx

    # get a pruned space, instead of 7 mem level just have 6 excluding the Register
    def get_factor_config_from_prune_idx(self, idx):
        factor_space = list(self.factor_space_tup)
        factor_space = [factor - 1 if factor > 1 else factor for factor in factor_space]
        factor_idx_arr = np.unravel_index(idx, tuple(factor_space))

        factor_config = copy.deepcopy(self.factor_space)
        idx = 0
        for i, sublist in enumerate(factor_config):
            for j, item in enumerate(sublist):
                val = factor_idx_arr[idx]
                if self.factor_space[i][j] > 1:
                    val += 1
                factor_config[i][j] = val
                idx += 1
        return factor_config

    def get_factor_config_from_idx(self, idx):
        factor_idx_arr = np.unravel_index(idx, self.factor_space_tup)
        factor_config = copy.deepcopy(self.factor_space)
        idx = 0
        for i, sublist in enumerate(factor_config):
            for j, item in enumerate(sublist):
                factor_config[i][j] = factor_idx_arr[idx]
                idx += 1
        return factor_config

    def get_idx_from_factor_config(self, factor_config):
        factor_config_tup = tuple([item for sublist in factor_config for item in sublist])
        idx = np.ravel_multi_index(factor_config_tup, self.factor_space_tup)
        return idx

    def get_default_factor(self):
        factor_config = []
        for prob_factor in self.unschduled_prob_factors:
            # init factor to temporal outer most loop 
            factors = [self.arch.mem_levels - 1] * len(prob_factor)
            factor_config.append(factors)
        return factor_config

    def get_default_perm(self):
        perm_config = []
        for mem_idx in range(self.arch.mem_levels):
            # for org_idx in range(self.mapspace.shape[1]):
            permutation = []
            for prob_idx in range(self.prob.prob_levels):
                permutation.append(prob_idx)
            perm_config.append(permutation)
        return perm_config

        # spatial_config is to make it backforward compatible

    # if spatial_configs = '0_0_0_0_0_0_0', no preset spatial dim for the NoC
    def reset_mapspace(self, spatial_config, spatial_configs=None):
        if spatial_configs is None:
            spatial_configs = [1] * self.prob.prob_levels
            prob_idx = self.prob.prob_name_idx_dict[spatial_config]  # C-4 K-5
            spatial_configs[prob_idx] = 16
        self._reset_mapspace(spatial_configs)

    def _reset_mapspace(self, spatial_configs):
        # reset mapspace
        self.mapspace = np.ones((self.arch.mem_levels, len(self.org_idx_dict.items()),
                                 self.prob.prob_levels, len(self.config_idx_dict.items())), dtype=np.int16)

        # reset copy of prob_factors  
        self.unschduled_prob_factors = copy.deepcopy(self.prob.prob_factors)

        # TODO load the maspace contraints to mapspace
        # init the spatial PE
        spatial_mem_idx = self.mapspace.shape[0] - 1
        for i in range(len(self.arch.mem_instances) - 1, -1, -1):
            if self.arch.mem_instances[i] == 1:
                spatial_mem_idx = i
        # only spatial partition at DRAM  
        org_idx = 0

        is_spatial_specified = False
        for prob_idx, spatial_factor in enumerate(spatial_configs):
            if spatial_factor > 0:
                self.mapspace[spatial_mem_idx][org_idx][prob_idx][1] = spatial_factor

                utils.update_prime_factors(self.unschduled_prob_factors, prob_idx, spatial_factor)
                is_spatial_specified = True

        # for each factor choose a memory level to be in
        # TODO not considering the spatial probabiliy
        # including spatial and temporal dims 
        # spatial - 1 for the NoC cores 
        # TODO make a control signal to enable or disable heuristic spatial level
        num_valid_spatial_levels = len(self.valid_spatial_levels)
        if is_spatial_specified:
            num_valid_spatial_levels = num_valid_spatial_levels - 1

        self.total_factor_choices = self.arch.mem_levels + num_valid_spatial_levels

        self.factor_space = []
        for prob_factor in self.unschduled_prob_factors:
            # set the bound to one, if the factor is 1
            if len(prob_factor) == 1 and prob_factor[0] == 1:
                bounds = [1]
            else:
                bounds = [self.total_factor_choices] * len(prob_factor)
            self.factor_space.append(bounds)
        self.factor_space_tup = tuple([item for sublist in self.factor_space for item in sublist])

    def update_mapspace(self, perm_config, factor_config):
        # update factors
        for prob_idx, mem_levels in enumerate(factor_config):
            for factor_idx, mem_idx in enumerate(mem_levels):
                if mem_idx < self.arch.mem_levels:
                    self.mapspace[mem_idx, 1, prob_idx, 1] *= self.unschduled_prob_factors[prob_idx][factor_idx]
                else:
                    spatial_mem_idx = self.valid_spatial_levels[mem_idx - self.arch.mem_levels][0]
                    self.mapspace[spatial_mem_idx, 0, prob_idx, 1] *= self.unschduled_prob_factors[prob_idx][factor_idx]
        # update permutation
        for mem_idx, permutations in enumerate(perm_config):
            for prob_idx, order in enumerate(permutations):
                self.mapspace[mem_idx, :, prob_idx, 0] = order

    def config_feature(self, perm_config, factor_config, org_idx=1):
        features = np.zeros((self.arch.mem_levels, self.prob.prob_levels), np.int16)
        for mem_idx in range(self.mapspace.shape[0]):
            for prob_idx in range(self.mapspace.shape[2]):
                factor = self.mapspace[mem_idx][org_idx][prob_idx][1]
                perm_idx = self.mapspace[mem_idx][org_idx][prob_idx][0]
                features[mem_idx][perm_idx] = factor
        logger.debug("features {}".format(features))
        return features

    def config_space_str(self, spatial_config, perm_config, factor_config):
        prob_factor_strs = []
        for mem_levels in factor_config:
            # factors = [self.arch.mem_levels-1] * len(prob_factor)
            # self.factor_config.append(factors)
            mem_level_arr = [str(mem_level) for mem_level in mem_levels]
            mem_level_str = "_".join(mem_level_arr)
            prob_factor_strs.append(mem_level_str)
        prob_factor_str = "-".join(prob_factor_strs)
        prob_perm_strs = []
        for mem_idx, permutations in enumerate(perm_config):
            perm_arr = [str(perm) for perm in permutations]
            perm_str = "".join(perm_arr)
            prob_perm_strs.append(perm_str)
        prob_perm_str = "-".join(prob_perm_strs)

        config_str = spatial_config + '+' + prob_perm_str + '+' + prob_factor_str
        # assert(len(config_str) < 256)
        return config_str

    def config_factor_str(self, org_idx):
        factor_strs = []
        for mem_idx in range(self.mapspace.shape[0]):
            factor_arr = []
            for prob_idx in range(self.mapspace.shape[2]):
                factor_arr.append(str(self.mapspace[mem_idx][org_idx][prob_idx][1]))
            factor_str = '_'.join(factor_arr)
            factor_strs.append(factor_str)
        config_factor_str = "-".join(factor_strs)
        return config_factor_str

    def config_perm_str(self, org_idx):
        perm_strs = []
        for mem_idx in range(self.mapspace.shape[0]):
            perm_arr = []
            for prob_idx in range(self.mapspace.shape[2]):
                perm_arr.append(str(self.mapspace[mem_idx][org_idx][prob_idx][0]))
            perm_str = "".join(perm_arr)
            perm_strs.append(perm_str)
        config_perm_str = "-".join(perm_strs)
        return config_perm_str

    def config_str(self):
        config_str = []
        for org_idx in range(self.mapspace.shape[1]):
            factor_strs = []
            config_factor_str = self.config_factor_str(org_idx)
            config_perm_str = self.config_perm_str(org_idx)
            config_str.append(config_factor_str + '+' + config_perm_str)
        return config_str

        # TODO Write fast valid mapping checker, instead of running timeloop every time

    def is_valid_mapping(self, factor_config, factor_space):
        # there should not be any factor allocated at the  the register level  
        for prob_idx, mem_levels in enumerate(factor_config):
            for mem_level, _ in enumerate(mem_levels):
                if (factor_config[prob_idx][mem_level] == 0) and (factor_space[prob_idx][mem_level] != 1):
                    return False
        return True

    # return valid spatial factor bound from the arch
    def get_valid_spatial_levels(self, use_valid_spatial_levels):
        valid_spatial_levels = []
        # for i, mem in enumerate(self.arch.storage):
        for i, instances in enumerate(self.arch.mem_instances):
            if i == 0:
                if not use_valid_spatial_levels:
                    valid_spatial_levels.append((i, 1))
                continue
            # inner_instances = self.arch.storage[i-1]['instances']
            inner_instances = self.arch.mem_instances[i - 1]
            # cur_instances = mem['instances']
            cur_instances = self.arch.mem_instances[i]
            assert (inner_instances % cur_instances == 0)
            spatial_factor = inner_instances // cur_instances
            if use_valid_spatial_levels:
                if spatial_factor != 1:
                    valid_spatial_levels.append((i, spatial_factor))
            else:
                valid_spatial_levels.append((i, spatial_factor))
        return valid_spatial_levels

    def random_factor_config(self):
        factor_config = []
        # for different prob dim, # factors choose a mem lepatial_levelsyel to be in 
        for prob_idx, mem_levels in enumerate(self.factor_space):
            factors = np.random.randint(self.total_factor_choices, size=len(mem_levels))
            factor_config.append(factors)
        return factor_config

    def generate_tile_str(self, factors):
        tile_arr = []
        for name, idx in self.prob.prob_name_idx_dict.items():
            tile_arr.append("{}={}".format(name, factors[idx]))
        tile_str = " ".join(tile_arr)
        return tile_str

    # permutations store the loop order of each var
    def generate_perm_str(self, permutations):
        perm_arr = [None] * len(permutations)
        for name, idx in self.prob.prob_name_idx_dict.items():
            perm_arr[permutations[idx]] = name

        # There should be no None if permutation is right 
        perm_str = "".join(perm_arr)
        return perm_str

    def get_input_related_mem(self):
        var_idx = self.var_idx_dict['Inputs']
        mems = []
        for mem_idx in range(self.arch.mem_levels):
            if self.bypass[mem_idx][var_idx] != 1:
                mems.append(mem_idx)
        return mems

    def get_mem_var(self):
        mem_var_dict = {}
        for mem_idx in range(self.arch.mem_levels):
            mem_var_dict[mem_idx] = []

        for mem_idx in range(self.arch.mem_levels):
            for k, v in self.var_idx_dict.items():
                if self.bypass[mem_idx][v] != 1:
                    mem_var_dict[mem_idx].append(k)
        return mem_var_dict

    def get_var_mem(self):
        var_mem_dict = {}
        for var in self.var_idx_dict.keys():
            var_mem_dict[var] = []

        for var, idx in self.var_idx_dict.items():
            for mem_idx in range(self.arch.mem_levels):
                if self.bypass[mem_idx][idx] != 1:
                    var_mem_dict[var].append(mem_idx)
        return var_mem_dict

    def is_mem_related_to_prob(self, mem_idx, prob_idx):
        related = False
        for var in self.prob_var_dict[prob_idx]:
            if mem_idx in self.var_mem_dict[var]:
                related = True
        return related

    def get_mem_util(self):
        var_size = len(self.var_idx_dict)
        utilized_mem_entries = np.ones((self.mapspace.shape[0], var_size), dtype=np.uint32)
        utilized_mem_prob_entries = np.ones((self.mapspace.shape[0], self.mapspace.shape[2], var_size), dtype=np.uint32)
        mem_prob_entries = np.ones((self.mapspace.shape[0], self.mapspace.shape[2]), dtype=np.uint32)
        # get utilized mem size, using both spatial and temporal dims
        for mem_idx in range(self.mapspace.shape[0]):
            for org_idx in range(self.mapspace.shape[1]):
                for prob_idx in range(self.mapspace.shape[2]):
                    # for each factor that is not zero
                    factor = self.mapspace[mem_idx][org_idx][prob_idx][1]
                    if factor > 1:
                        for var in self.prob_var_dict[prob_idx]:
                            var_idx = self.var_idx_dict[var]
                            for utilized_mem_idx in self.var_mem_dict[var]:
                                if utilized_mem_idx >= mem_idx:
                                    utilized_mem_entries[utilized_mem_idx][var_idx] *= factor
                                    utilized_mem_prob_entries[utilized_mem_idx][prob_idx][var_idx] *= factor

                        for utilized_mem_idx in range(self.mapspace.shape[0]):
                            if utilized_mem_idx >= mem_idx:
                                mem_prob_entries[utilized_mem_idx][prob_idx] *= factor
        logger.debug("utilized_mem_prob_entries: {}".format(utilized_mem_prob_entries))
        # For buffers that are related to P and Q we need to see if R and S is 
        # TODO hardcoded here. for the input buffer, or buffers stores Inputs from the bypass constraints 
        # I = (P - 1) * Stride + R  
        input_related_mems = self.get_input_related_mem()
        input_idx = self.var_idx_dict["Inputs"]
        for input_related_mem_idx in input_related_mems:
            # for dram assume no extra input needs to be loaded
            if utilized_mem_prob_entries[input_related_mem_idx][2][input_idx] > 1 or \
                    mem_prob_entries[input_related_mem_idx][0] > 1:
                utilized_mem_prob_entries[input_related_mem_idx][2][input_idx] = (mem_prob_entries[
                                                                                      input_related_mem_idx][2] - 1) * \
                                                                                 self.prob.prob['Wstride'] + \
                                                                                 mem_prob_entries[
                                                                                     input_related_mem_idx][0]
                strid = self.prob.prob['Wstride']
            if utilized_mem_prob_entries[input_related_mem_idx][3][input_idx] > 1 or \
                    mem_prob_entries[input_related_mem_idx][1] > 1:
                utilized_mem_prob_entries[input_related_mem_idx][3][input_idx] = (mem_prob_entries[
                                                                                      input_related_mem_idx][3] - 1) * \
                                                                                 self.prob.prob['Hstride'] + \
                                                                                 mem_prob_entries[
                                                                                     input_related_mem_idx][1]
            # Update entry factors 
            # logger.debug("utilized_mem_entries before {}".format(utilized_mem_entries[input_related_mem_idx][input_idx]))
            utilized_mem_entries[input_related_mem_idx][input_idx] = 1
            for prob_idx in range(self.mapspace.shape[2]):
                utilized_mem_entries[input_related_mem_idx][input_idx] *= \
                    utilized_mem_prob_entries[input_related_mem_idx][prob_idx][input_idx]
            # logger.debug("utilized_mem_entries after {}".format(utilized_mem_entries[input_related_mem_idx][input_idx]))

        # logger.debug("utilized_mem_prob_entries after: {}".format(utilized_mem_prob_entries))

        utilized_mem_instances = [1] * self.mapspace.shape[0]
        utilized_spatial_factors = [1] * self.mapspace.shape[0]

        prev_spatial_instances = 1
        # for instances, check spatial dimension only 
        for mem_idx in range(self.mapspace.shape[0] - 1, -1, -1):
            org_idx = 0
            utilized_mem_instances[mem_idx] = prev_spatial_instances
            for prob_idx in range(self.mapspace.shape[2]):
                # for each factor that is not zero
                factor = self.mapspace[mem_idx][org_idx][prob_idx][1]
                if factor > 1:
                    prev_spatial_instances *= factor
                    utilized_spatial_factors[mem_idx] *= factor

        logger.debug("utilized_mem_entries: {}".format(utilized_mem_entries))
        logger.debug("utilized_spatial_factors: {}".format(utilized_spatial_factors))
        return utilized_mem_entries, utilized_mem_instances, utilized_spatial_factors

    def valid_check(self):
        utilized_mem_entries, utilized_mem_instances, utilized_spatial_factors = self.get_mem_util()

        # sum up all var mem util size for each mem level 
        total_utilized_mem_entries = np.zeros(self.mapspace.shape[0], dtype=np.uint32)
        for mem_idx in range(self.arch.mem_levels - 1):
            for var_name, var_idx in self.var_idx_dict.items():
                if mem_idx in self.var_mem_dict[var_name]:
                    total_utilized_mem_entries[mem_idx] += utilized_mem_entries[mem_idx][var_idx]

        utilized_mem_entries = total_utilized_mem_entries
        logger.debug("summed utilized_mem_entries: {}".format(utilized_mem_entries))

        valid = True
        invalid_mem_entries = []
        invalid_mem_instances = []
        logger.debug("utilized_mem_instances: {}".format(utilized_mem_instances))
        logger.debug("mem_entries: {}".format(self.arch.mem_entries))
        logger.debug("mem_instances: {}".format(self.arch.mem_instances))
        logger.debug("valid_spatial_factors: {}".format(self.valid_spatial_factors))

        # check if total entries/instances exceeds the maximum size/instances, 
        # except for DRAM
        for i, mem_entries in enumerate(utilized_mem_entries):
            if utilized_mem_instances[i] > self.arch.mem_instances[i] or utilized_spatial_factors[i] > \
                    self.valid_spatial_factors[i]:
                valid = False
                invalid_mem_instances.append(i)
            if i < len(utilized_mem_entries) - 1 and mem_entries > self.arch.mem_entries[i]:
                valid = False
                invalid_mem_entries.append(i)

        logger.debug("invalid_mem_entries: {}".format(invalid_mem_entries))
        logger.debug("invalid_mem_instances: {}".format(invalid_mem_instances))
        return (valid, invalid_mem_entries, invalid_mem_instances, utilized_mem_entries, utilized_mem_instances)

    def generate_greedy_mapspace(self):
        mapping = {}
        mapping['mapping'] = []
        map_prob_bound = [1] * self.mapspace.shape[2]

        for mem_idx in range(self.mapspace.shape[0]):
            for org_idx in range(self.mapspace.shape[1]):
                factors = []
                permutations = []
                config = {}
                # No fanout to inside mem for the innermost mem
                if org_idx == 0:
                    # if mem_idx < self.mapspace.shape[0] - 1:
                    if mem_idx > 0:
                        if self.arch.storage[mem_idx]['instances'] == self.arch.storage[mem_idx - 1]['instances']:
                            continue
                    elif mem_idx == 0:
                        continue

                config['target'] = self.arch.mem_name[mem_idx]
                config['type'] = self.org_name_dict[org_idx]
                for prob_idx in range(self.mapspace.shape[2]):
                    permutations.append(self.mapspace[mem_idx][org_idx][prob_idx][0])
                    factors.append(self.mapspace[mem_idx][org_idx][prob_idx][1])
                    if self.mapspace[mem_idx][org_idx][prob_idx][1] != 1:
                        map_prob_bound[prob_idx] *= self.mapspace[mem_idx][org_idx][prob_idx][1]
                config['factors'] = self.generate_tile_str(factors)
                config['permutation'] = self.generate_perm_str(permutations)
                mapping['mapping'].append(config)

        # Make sure the prob size for the mapping matches the orig prob size
        try:
            assert (map_prob_bound == self.prob.prob_bound[0:len(map_prob_bound)])
        except:
            logger.error("{} not equals to {}".format(map_prob_bound, self.prob.prob_bound[0:len(map_prob_bound)]))

        for mem_idx in range(self.bypass.shape[0]):
            bypass = []
            keep = []
            config = {}
            config['target'] = self.arch.mem_name[mem_idx]
            config['type'] = 'bypass'
            for var_idx in range(self.bypass.shape[1]):
                if self.bypass[mem_idx][var_idx]:
                    bypass.append(self.var_name_dict[var_idx])
                else:
                    keep.append(self.var_name_dict[var_idx])
            config['bypass'] = bypass
            config['keep'] = keep
            mapping['mapping'].append(config)

        return mapping

    def generate_greedy_mapspace(self, template_dict):
        map_prob_bound = [1] * self.mapspace.shape[2]

        for mem_idx in range(self.mapspace.shape[0]):
            for org_idx in range(self.mapspace.shape[1]):
                factors = []
                config = {}
                # No fanout to inside mem for the innermost mem
                if org_idx == 0:
                    # if mem_idx < self.mapspace.shape[0] - 1:
                    if mem_idx > 0:
                        if self.arch.storage[mem_idx]['instances'] == self.arch.storage[mem_idx - 1]['instances']:
                            continue
                    elif mem_idx == 0:
                        continue

                config['target'] = self.arch.mem_name[mem_idx]
                config['type'] = self.org_name_dict[org_idx]
                for prob_idx in range(self.mapspace.shape[2]):
                    factors.append(self.mapspace[mem_idx][org_idx][prob_idx][1])
                    if self.mapspace[mem_idx][org_idx][prob_idx][1] != 1:
                        map_prob_bound[prob_idx] *= self.mapspace[mem_idx][org_idx][prob_idx][1]
                config['factors'] = self.generate_tile_str(factors)
                template_dict['mapspace']['constraints'].append(config)

        # Make sure the prob size for the mapping matches the orig prob size
        try:
            assert (map_prob_bound == self.prob.prob_bound[0:len(map_prob_bound)])
        except:
            logger.error("{} not equals to {}".format(map_prob_bound, self.prob.prob_bound[0:len(map_prob_bound)]))

    def generate_mapping(self):
        mapping = {}
        mapping['mapping'] = []
        map_prob_bound = [1] * self.mapspace.shape[2]

        for mem_idx in range(self.mapspace.shape[0]):
            for org_idx in range(self.mapspace.shape[1]):
                factors = []
                permutations = []
                config = {}
                # No fanout to inside mem for the innermost mem
                if org_idx == 0:
                    # if mem_idx < self.mapspace.shape[0] - 1:
                    if mem_idx > 0:
                        # if self.arch.storage[mem_idx]['instances'] == self.arch.storage[mem_idx-1]['instances']:
                        if self.arch.mem_instances[mem_idx] == self.arch.mem_instances[mem_idx - 1]:
                            continue
                    elif mem_idx == 0:
                        continue

                config['target'] = self.arch.mem_name[mem_idx]
                config['type'] = self.org_name_dict[org_idx]
                for prob_idx in range(self.mapspace.shape[2]):
                    permutations.append(self.mapspace[mem_idx][org_idx][prob_idx][0])
                    factors.append(self.mapspace[mem_idx][org_idx][prob_idx][1])
                    if self.mapspace[mem_idx][org_idx][prob_idx][1] != 1:
                        map_prob_bound[prob_idx] *= self.mapspace[mem_idx][org_idx][prob_idx][1]
                config['factors'] = self.generate_tile_str(factors)
                config['permutation'] = self.generate_perm_str(permutations)
                mapping['mapping'].append(config)

        # Make sure the prob size for the mapping matches the orig prob size
        try:
            assert (map_prob_bound == self.prob.prob_bound[0:len(map_prob_bound)])
        except:
            logger.error("{} not equals to {}".format(map_prob_bound, self.prob.prob_bound[0:len(map_prob_bound)]))

        for mem_idx in range(self.bypass.shape[0]):
            bypass = []
            keep = []
            config = {}
            config['target'] = self.arch.mem_name[mem_idx]
            config['type'] = 'bypass'
            for var_idx in range(self.bypass.shape[1]):
                if self.bypass[mem_idx][var_idx]:
                    bypass.append(self.var_name_dict[var_idx])
                else:
                    keep.append(self.var_name_dict[var_idx])
            config['bypass'] = bypass
            config['keep'] = keep
            mapping['mapping'].append(config)

        return mapping

    def print(self):
        print(self.__dict__)


class Mapping(object):
    """ Hardware Mapping"""

    def __init__(self):
        temporal = []
        spatial = []
        bypass = []


def gen_all_config(run, pickle_path=None, status_file=None, run_gen_map=True, run_gen_tc=True, run_sim_test=True):
    prob_path = pathlib.Path('timeloop_configs/01_3x3Conv/prob/toy_prob.yaml').resolve()
    prob = Prob(prob_path)

    arch_path = pathlib.Path('timeloop_configs/arch/simba.yaml').resolve()
    arch = Arch(arch_path)

    mapspace_path = pathlib.Path('timeloop_configs/00_1x1Conv/mapspace/mapspace.yaml').resolve()
    mapspace = Mapspace(mapspace_path)

    # init
    mapspace.init(prob, arch)

    # try to iterate from valid mapping
    # for prob_idx, mem_levels in enumerate(mapspace.factor_space):
    # for mem_level in mem_levels: # specify the mem_level of the current factor
    def is_config_done(factor_config, factor_space):
        prob_idx = len(factor_space) - 1
        mem_level = len(factor_space[prob_idx]) - 1
        if factor_config[prob_idx][mem_level] < factor_space[prob_idx][mem_level]:
            return False
        else:
            return True

    def config_increment(factor_config, factor_space, factor_org_idx):
        loop_carry = True
        for i, idx_tup in enumerate(factor_org_idx):
            prob_idx, mem_level = idx_tup
            # specify the mem_level of the current factor 
            if loop_carry:
                new_config = factor_config[prob_idx][mem_level] + 1
            if new_config < factor_space[prob_idx][mem_level]:
                loop_carry = False
                factor_config[prob_idx][mem_level] = new_config
                break
            else:
                if ((prob_idx == len(mapspace.factor_space) - 1) and
                        (mem_level == len(mem_levels) - 1)):
                    factor_config[prob_idx][mem_level] = new_config
                else:
                    factor_config[prob_idx][mem_level] = 0

    def get_factor_org_idx(factor_space):
        factor_org_idx = []
        for prob_idx, mem_levels in enumerate(factor_space):
            for mem_level, _ in enumerate(mem_levels):
                factor_org_idx.append((prob_idx, mem_level))
        return factor_org_idx

    # Compile simulation framework
    if run:
        x_size = 4
        y_size = 4
        utils.gen_makefile(x_size, y_size)
        utils.compile_sim()

    # key: config_space_str
    # val: [idx, run_gen_map, run_gen_tc, run_sim_cycle]
    # -1 means didnt run, positive value indicates different info
    # is_valid_mapping is determined with is_valid_mapping and timeloop run
    # sim_cycle: return by sim_test 
    if status_file:
        status_file = pathlib.Path(status_file)
        if status_file.is_file():
            status_dict = utils.parse_json(status_file)
    else:
        status_dict = {}

    configs = []
    if pickle_path:
        logger.info(pickle_path)

    perm_config = mapspace.get_default_perm()
    for spatial_config in mapspace.spatial_space:
        mapspace.reset_mapspace(spatial_config)
        factor_config = mapspace.get_default_factor()
        for prob_idx, mem_levels in enumerate(mapspace.factor_space):
            for mem_level, _ in enumerate(mem_levels):  # specify the mem_level of the current factor
                factor_config[prob_idx][mem_level] = 0
        logger.info("Init Iter: {} in {}".format(factor_config, mapspace.factor_space))

        factor_org_idx = get_factor_org_idx(mapspace.factor_space)
        idx = 0
        while (not is_config_done(factor_config, mapspace.factor_space)):

            if not mapspace.is_valid_mapping(factor_config, mapspace.factor_space):
                config_increment(factor_config, mapspace.factor_space, factor_org_idx)
                continue
            if run:
                logger.info("Iter: {} in {}".format(factor_config, mapspace.factor_space))
                if status_file:
                    run_config.run_config(mapspace, spatial_config, perm_config, factor_config, status_dict,
                                          run_gen_map, run_gen_tc, run_sim_test)
                else:
                    run_config.run_config(mapspace, spatial_config, perm_config, factor_config, dict(), run_gen_map,
                                          run_gen_tc, run_sim_test)
            # if pickle_path is defined, store the pickle config
            if pickle_path:
                configs.append((copy.deepcopy(spatial_config), copy.deepcopy(perm_config), copy.deepcopy(factor_config),
                                run_gen_map, run_gen_tc, run_sim_test))
            config_increment(factor_config, mapspace.factor_space, factor_org_idx)
            utils.delete_dramsim_log()
            idx += 1
    if status_file:
        utils.store_json(status_file, status_dict)

    if pickle_path:
        utils.store_pickle(pickle_path, configs)


if __name__ == "__main__":
    module_name = pathlib.Path(__file__).stem
    utils.setup_logging(module_name, logger)

    # unit test
    # assert(test_run())
    config_pkl = pathlib.Path('config.pkl')
    if not config_pkl.is_file():
        gen_all_config(False, pickle_path=config_pkl)
    configs = utils.parse_pickle(config_pkl)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cosa']

package_data = \
{'': ['*'],
 'cosa': ['configs/arch/*',
          'configs/mapspace/*',
          'configs/workloads/alexnet_graph/*',
          'configs/workloads/alexnet_pad_graph/*',
          'configs/workloads/deepbench_graph/*',
          'configs/workloads/densenet161_graph/*',
          'configs/workloads/dlrm_graph/*',
          'configs/workloads/resnet50_graph/*',
          'configs/workloads/resnext50_32x4d_graph/*',
          'configs/workloads/vgg16_graph/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'gurobipy>=9.5.1,<10.0.0', 'numpy>=1.22.3,<2.0.0']

entry_points = \
{'console_scripts': ['cosa = cosa.cosa:run_cosa']}

setup_kwargs = {
    'name': 'cosa-scheduler',
    'version': '0.1.0',
    'description': 'A constrained-optimization based scheduler for spatial DNN accelerators',
    'long_description': 'None',
    'author': 'Qijing Huang',
    'author_email': 'qijing.huang@berkeley.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
